// Author: Pete Kirkham
//
// quad store matching query bench.
//
#include <xmmintrin.h>
#include <iostream>

#define KiB (1024)
#define MiB (1024*1024)
#define GiB (1024*1024*1024)
#define BLOCK_SIZE (16*KiB)

#include "utils/string_utils.h"


__m128i* match2 (int value, __m128i* data, __m128i* result, size_t size) ;
void dump (__m128i* data, size_t limit) ;
void ndump (__m128i* data, size_t limit) ;


int main (int narg, char** argvp)
{
    size_t num_quads(1 * MiB);

    if (narg > 1)
    {
        num_quads = atoi(argvp[1]) * MiB;
    }

    size_t num_passes(16);

    if (narg > 2)
    {
        num_passes = atoi(argvp[2]);
    }

    std::cout << "number of quads: " << num_quads << std::endl;
    std::cout << "memory required: " << (2 * num_quads * 16) / MiB << " MiB" << std::endl;

    // create num_quads quads
    __m128i* data = new __m128i[num_quads];
    __m128i* result = new __m128i[num_quads];
    __m128i* result_tail(result);

    // fill in data
    for (size_t i(0); i < num_quads; ++i)
    {
        _mm_storeu_si128(data+i, _mm_set_epi32((i&0xff)^0x34, ((i>>2)&0xff)^0xd0, ((i>>4)&0xff)^0x52, 0xffffffff));
        _mm_storeu_si128(result+i, _mm_set_epi32(0xffffffff, 0xffffffff, 0xffffffff, 0));
    }

    // dump first 16 data
    // dump(data, 16);

    // match data
    for (size_t pass(0); pass < num_passes; ++pass)
    {
        result_tail = match2(210, data, result, num_quads);
        result_tail = match2(211, data, result_tail, num_quads);
        result_tail = match2(212, data, result_tail, num_quads);
        result_tail = match2(213, data, result_tail, num_quads);
        std::cout << "pass " << pass << std::endl;
    }

    // output result
    size_t nresults(size_t(result_tail - result));

    if (nresults > 16)
        std::cout << nresults << " results, first 16 printed: " << std::endl;
    else
        std::cout << nresults << " results: " << std::endl;

    ndump(result, 16);

    delete [] result;
    delete [] data;

    std::cout << "done." << std::endl;
}

// matches the second value in quad
__m128i* match2 (int value, __m128i* data, __m128i* result, size_t size)
{
    __m128i mask(_mm_set_epi32(0, value, 0, 0));
/*
    std::cout << "mask   " << to_i4(mask) << std::endl;
    {
        __m128i in(_mm_set_epi32(1,2,3,4));
        std::cout << "in     " << to_i4(in) << std::endl;
        std::cout << "cmpeq(in, mask) => " << to_i4(_mm_cmpeq_epi32(in, mask)) << std::endl;
        std::cout << "spread(..) => " << to_i4(_mm_shuffle_epi32(_mm_cmpeq_epi32(in, mask), SPREAD_EPI32(1))) << std::endl;
        std::cout << "and(in, ..) => " << to_i4(_mm_and_si128(in, _mm_shuffle_epi32(_mm_cmpeq_epi32(in, mask), SPREAD_EPI32(1)))) << std::endl;
        std::cout << std::endl;
    }

    {
        __m128i in(_mm_set_epi32(1,value,3,4));
        std::cout << "in     " << to_i4(in) << std::endl;
        std::cout << "cmpeq(in, mask) => " << to_i4(_mm_cmpeq_epi32(in, mask)) << std::endl;
        std::cout << "spread(..) => " << to_i4(_mm_shuffle_epi32(_mm_cmpeq_epi32(in, mask), SPREAD_EPI32(1))) << std::endl;
        std::cout << "and(in, ..) => " << to_i4(_mm_and_si128(in, _mm_shuffle_epi32(_mm_cmpeq_epi32(in, mask), SPREAD_EPI32(1)))) << std::endl;
        std::cout << std::endl;
    }
*/

    for (size_t i(0); i < size; ++i)
    {
        __m128i in(_mm_loadu_si128(data+i));
        __m128i cmp(_mm_shuffle_epi32(_mm_cmpeq_epi32(in, mask), SPREAD_EPI32(1)));
        _mm_storeu_si128(result, _mm_and_si128(in, cmp));
        result += _mm_extract_epi16(cmp, 0) & 1;
    }

    return result;
}

// dumps the data
void dump (__m128i* data, size_t limit)
{
    for (size_t i(0); i < limit; ++i)
        std::cout << to_i4(_mm_loadu_si128(data+i)) << std::endl;
}

// dumps the data until first null
void ndump (__m128i* data, size_t limit)
{
    for (size_t i(0); i < limit; ++i)
        if (!(*((int*)(&data[i]))))
            break;
        else
            std::cout << to_i4(_mm_loadu_si128(data+i)) << std::endl;
}
